		 /* Copyright 1993 Ed Casas  */

#define Version		  "efax v 0.5"

#include <ctype.h>	    /* ANSI C */
#include <errno.h>
#include <signal.h>    
#include <stdarg.h> 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <fcntl.h>	    /* UNIX */
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include <sys/stat.h>       /* required for SYSV (?) */


/* constants... */

#ifndef FILENAME_MAX
#define FILENAME_MAX 255
#endif

#define FAXDEV "/dev/fax"   /* default fax modem device */

#define LOGF     stderr     /* session log written to this stream */

			    /* delays, milliseconds */
#define T_CMD     200	    /* delay before each modem command */
#define T_DTR     200	    /* duration of DTR-off pulse to reset modem */
#define T_ESC     3500	    /* delay before ``Hayes'' escape sequence */

			    /* timeouts, seconds */
#define TO_RESET  2	    /* during modem resync */
#define TO_INIT   10	    /* local modem setup */
#define TO_DATAF  10	    /* software adaptive answer data connect */
#define TO_A      120	    /* dial or answer (Phase A) - modem may TO first */
#define TO_B      30	    /* local<->remote handshaking (Phase B) */
#define TO_C      15	    /* waiting for data (Phase C) */

#define CMDCBAUD  B19200    /* default baud rate flag for data & modem init */
#define CMDBUFSIZE 128      /* longest possible modem command or response */
#define DEFCAP 1,3,0,2,0,0,0,0	/* default local capabilities */
#define FNAMFT "%m%d%H%M%S" /* strftime() format for default file name */
#define HDBLKFLAG '#'	    /* prefix to force HDB (text) lock file style */
#define LOCKPOLLDELAY 15    /* seconds between checks of lock files */
#define MAXGETTY 512        /* maximum length of ``getty'' (-g) command */
#define MAXIOPT  100        /* maximum # of modem initialization commands */
#define MAXLKFILE 16	    /* maximum number of lock files */
#define MAXNULLS 2	    /* maximum consecutive received nulls saved */
#define MAXTSTAMP 80	    /* maximum length of a time stamp */
#define MINWRITE  32        /* minimum bytes per write() */
#define NCAP 8              /* number of fields in a capability string */
#define NTXRETRY  3	    /* maximum re-sends per page */
#define RCVBUFSIZE 1024	    /* read up to this many bytes at a time from fax */
#define SNDBUFSIZE 1024	    /* maximum bytes to write at a time to fax */
#define T4RTCLEN  9	    /* end of page mark for T.4 coded images */
#define T4RTC    "\000\020\001\000\020\001\000\020\001"

			    /* signals to be caught so can hang up phone */
#define SIGNALS  SIGHUP, SIGINT, SIGQUIT, SIGIOT, SIGSEGV, SIGALRM, SIGTERM 


const char *Usage =
  "Usage:\n"
  "  %s [ option ]... [ -r pat | -t num file... ]\n"
"Options:\n"
  "  -b bps  use bps baud rate to initialize modem (default 19200)\n"
  "  -c cap  set file format or receive capabilites to cap\n"
  "  -d dev  use modem on device dev (- for stdin/out, default " FAXDEV ")\n"
  "  -g cmd  exec /bin/sh -c cmd for data calls\n"
  "  -i str  send modem command ATstr\n"
  "  -l id   set local indetification to id\n"
  "  -o opt  use protocol option opt:\n"
/*"      1     use class 1 modem commands\n"*/
/*"      2     use class 2 modem commands\n"*/
  "      a     if initial (data mode) answer fails, retry as fax\n"
  "      b     use 38.4 kbps on modem port\n"
  "      e     ignore errors in modem initialization commands\n"
  "      f     don't use xon/xoff flow control when receiving\n"
  "      h     use hardware (RTS/CTS) flow control when transmitting\n"
  "      p     pad transmitted lines or remove received padding\n"
  "      r     reverse bit order\n"  
  "      s     use 2 stop bits\n"  
  "      x     use XON instead of DC2 to trigger reception\n"
  "      w     don't wait for trigger character before sending\n"
  "      z     add 200 ms to pause before modem comands (cumulative)\n"
  "  -s      share (unlock) modem device while waiting for call\n"
  "  -w      wait for OK or CONNECT prompt instead of answering phone\n"
  "  -v      display modem commands and responses (-v again for more)\n"
  "  -x fil  use uucp-style lock file fil\n"
  "Commands:\n"
  "  -r      answer and receive fax into files pat.001, pat.002, ... \n"
  "  -t      send fax image files file... to telephone num\n"
  ;

enum  cchar {				/* control characters */
  NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS,  HT,  LF,
  VT,  FF,  CR,  SO,  SI,  DLE, XON, DC2, XOFF,DC4, NAK,
  SYN, ETB, CAN, EM,  SUB, ESC, FS,  GS,  RS,  US } ;

const char *ccname [ 32 ] = {		/* control character names */
  "NUL","SOH","STX","ETX","EOT","ENQ","ACK","BEL","BS",  "HT", "LF",
  "VT", "FF", "CR", "SO", "SI", "DLE","XON","DC2","XOFF","DC4","NAK",
  "SYN","ETB","CAN","EM", "SUB","ESC","FS", "GS", "RS",  "US" } ;

const char *responses[] = {		/* modem responses */
  "OOK", "O0", "CCONNECT", "C1", "NNO CARRIER", "N3", "EERROR","E4", "D5",
  "NNO DIALTONE" "N6", "BBUSY", "B7", "NNO ANSWER", "N8", 0 } ;

enum prompts {				/* codes for modem prompts */
   BUSY = 'B', CONNECT = 'C', OK = 'O', RING = 'R' } ;

enum ttymodes				/* serial port modes */
  { HUP_MODE, CMD_MODE , FAX_MODE, RX_MODE, TX_MODE, OLD_MODE } ;
					/* max. values of capability fields */
const int capmax [ NCAP ] = { 1, 5, 4, 2, 3, 2, 1, 7 } ;
					/* characters per second for br */
const int cps [ 6 ] = { 240, 480, 720, 960, 1200, 1440 } ;
					/* minimum scan time in ms  */
const int delay [ 8 ] = { 0 , 5, 10, 10, 20, 20, 40, 40 } ;

					/* capabilities stored in struct */
typedef struct { unsigned char vr, br, wd, ln, df, ec, bf, st ; } cap ;

/* program options. set in main(). */

char *argv0 = "?" ;			/* program name (argv[0]) */
char *getty = "" ;			/* command for data calls */
char *iopt[ MAXIOPT ] ;			/* modem initialization commands */
char *lkfiles [ MAXLKFILE+1 ] ;		/* lock files  */
char *localid = "" ;			/* local ID (telephone #) */
char startchar = DC2 ;			/* character to start reception  */
int c1 = 0 ;				/* use class 1 protocol */
int datafirst = 0 ;			/* use software adaptive answer */
int cmdcbaud = CMDCBAUD ;		/* data & modem init baud rate flag */
int flowctlrcv = 1 ;			/* use flow control on receive */
int hispeed = 0 ;			/* use 38.4 kbps on DTE-DCE */
int igniniterr=0 ;			/* ignore initialization errors */
int niopt=0 ;				/* number of initialization commands */
int padding = 0 ;			/* add padding on tx, strip on rx */
int cmdpause = T_CMD ;			/* delay before each command (ms) */
int rtscts = 0 ;			/* use RTS/CTS flow control */
int share = 0 ;		                /* share fax device */
int stopbits = 0 ;			/* use 2 stopbits */
int verb = 0 ;				/* verbose mode */
int wait2send = 1 ;			/* wait for DC2 or XON to send data */
int waitforcall = 0 ;			/* wait for incoming call */

unsigned char                           /* bit reversal lookup tables*/
  reversebits [ 256 ] , normalbits [ 256 ] , *bitorder = normalbits ;

					/* tables to find T.4 end-of-line */
unsigned char eolones [ 256 ] , eolprezero [ 256 ] , eolpostzero [ 256 ] ;

/* fax stream variables. */

char faxibuf [ RCVBUFSIZE ] ,		/* fax input stream. */
  *faxip=faxibuf , *faxiq=faxibuf ;

char faxobuf [ SNDBUFSIZE ] ,		/* fax output stream. */
  *faxop=faxobuf , *faxoq=faxobuf+SNDBUFSIZE ;

int faxin=-1, faxout=-1 ;		/* fax device file descriptors. */

/* session state variables. set in cmd() */

char lastprompt [ CMDBUFSIZE ] ;	/* last modem response (a prompt) */
int decimate = 0 ;			/* drop 1/2 lines to get 98 lpi */
int good=0 ;				/* last page transmitted OK */
int hsc=-1 ;				/* hangup status code */
int minlen = 0 ;			/* pad lines to this many characters */

cap local = { DEFCAP } ;		/* capabilities  */
cap remote, session ;			


/* Functions... */

/* Print time stamp. */

void tstamp( )
{
  static time_t start = 0 , now ;
  char tbuf [ MAXTSTAMP ] ;

  if ( ( now = time ( 0 ) ) - start > 60 ) {
    strftime ( tbuf , MAXTSTAMP , " (%c)" , localtime( &now ) ) ;
    start = now ;
  } else {
    sprintf ( tbuf , " (%d)" , now - start ) ;
  }
  fputs ( tbuf , LOGF ) ;
}


/* For systems without strerror(3) */

char *strerror( int i )       
{
  extern int sys_nerr;
  extern char *sys_errlist[];
  return ( i >= 0 && i < sys_nerr ) ? sys_errlist[i] : "Unknown Error" ;
}


/* Print a character or string to LOGF to log exchanges with the modem,
   showing non-printing characters.  */

void logc ( char c ) 
{
  if ( c < 0 || c > '\176' ) fprintf ( LOGF , "<%02x>", c ) ; 
  else if ( c >= ' ' ) putc ( c , LOGF ) ;
  else if ( verb >= 2 ) fprintf ( LOGF , "<%s>" , ccname [ (int) c ] ) ;
}

void logs ( char *p ) 
{
  for ( ; p && *p ; p++ ) logc ( *p ) ;
  if ( verb > 1 ) tstamp( ) ;
  putc ( '\n' , LOGF ) ;
  fflush ( LOGF ) ;
} 

/* Print a log message with a variable number of printf()-type
   arguments.  The message may contain a prefix of upper-case
   characters and digits for additional actions. P, E, or W
   expand into strings, S prints the error message for the most
   recent system error, and a digit sets the return value.
   Returns 0 if no prefix digits. */

int msg (char *s, ... ) 
{ 
  char err=0, *ps="", *pe="", *pw="" ;
  va_list ap ;
  va_start ( ap, s ) ;

  for ( ; *s && ( isupper (*s) || isdigit (*s) ) ; s++ ) {
    if ( *s == 'S' ) ps = strerror ( errno ) ;
    if ( *s == 'E' ) pe = "Error: " ;
    if ( *s == 'W' ) pw = "Warning: " ;
    if (isdigit(*s)) err = *s - '0' ;
  }

  fprintf ( LOGF , "%s: %s%s" , argv0 , pe , pw ) ;
  vfprintf( LOGF, s , ap ) ;
  fputs ( ps , LOGF ) ;
  tstamp ( ) ; 
  fputs ( "\n" , LOGF ) ;
  fflush ( LOGF ) ;

  va_end ( ap ) ;
  return err ;
}


/* Set serial port mode. Get current attributes and turn off everything
   that we might have to change.  Turn on bits corresponding to required
   mode and set the new attributes. Returns 0 or 2 on error. */

int ttymode ( enum ttymodes mode )
{
  int err=0, cbaud ;         
  static int saved=0 ;
  struct termios t ;  
  static struct termios orig ;

  if ( tcgetattr ( faxin , &t ) ) 
    err = msg("SE2can't get modem mode:") ;
  
  if ( ! err && ! saved ) { orig = t ; saved = 1 ; }

  t.c_iflag &= ~( IGNBRK | BRKINT | IGNPAR | INPCK | ISTRIP | INLCR | IGNCR
		 | ICRNL | IUCLC | IXON | IXANY | IXOFF | IMAXBEL ) ;
  t.c_oflag &= ~( OPOST ) ;
  t.c_cflag &= ~( CBAUD | CSIZE | CSTOPB | PARENB | HUPCL | CIBAUD | CRTSCTS );
  t.c_lflag &= ~( ISIG | ICANON | ECHO | NOFLSH | TOSTOP | FLUSHO | PENDIN ) ;
  t.c_cc[VMIN]  = 1 ; 
  t.c_cc[VTIME] = 0 ; 

  cbaud = ( mode == CMD_MODE ? cmdcbaud : ( hispeed ? B38400 : B19200 ) ) ;
  
  t.c_cflag |= cbaud | CS8 | CREAD | CLOCAL | ( stopbits ? CSTOPB : 0 ) ;

  if ( ! err ) 
    switch ( mode ) {
    case HUP_MODE : 
      t.c_cflag &= ~CBAUD ; 
      break ;
    case CMD_MODE : 
    case FAX_MODE : 
      break ;
    case  RX_MODE : 
      t.c_iflag |= flowctlrcv ? IXOFF : 0 ; 
      break ;
    case  TX_MODE : 
      t.c_iflag |= rtscts ? 0 : IXON ;
      t.c_cflag |= rtscts ? CRTSCTS : 0 ; 
      break ;
    case OLD_MODE : 
      if ( saved ) t = orig ; 
      break ;
    default       :  
      err = msg("E4can't happen(ttymode)") ; 
      break ;
    }
  
  if ( ! err && tcsetattr ( faxin , TCSADRAIN , &t ) )
    err = msg ( "SE2unable to set terminal mode:" ) ;
  
  return err ;
}


/* Return termios CBAUD value for speed s or EOF on error. */

int cbaud ( char *s )
{
  int i, b=-1, f=EOF ;

  int brate[] = { 50, B50, 75, B75, 110, B110, 134, B134, 150,
      B150, 200, B200, 300, B300, 600, B600, 1200, B1200, 1800,
      B1800, 2400, B2400, 4800, B4800, 9600, B9600, 19200,
      B19200, 38400, B38400, 0, 0 } ;

  sscanf ( s , "%d" , &b ) ;
  for ( i=0 ; brate[i] ; i+=2 ) if ( brate[i] == b ) { f=brate[i+1] ; break ; }
  if ( f == EOF ) msg ( "E2unknown baud rate (%s)" , s )  ;
  return f ;
}


/* faxgetc is a macro like getc().  It returns the next character from the
   fax device or EOF after t seconds of idle time. */

#define faxgetc(t) ( faxip >= faxiq && faxundrflw(t) == EOF ? EOF : \
		    *(unsigned char*)faxip++ )


/* Return number of characters ready to read or < 0 on error.  t is number
   of seconds of idle time before timing out.  If t negative, waits
   forever. */

int faxdata ( int t )
{
  int n ;
  fd_set fds ;

  struct timeval timeout ;

  timeout.tv_sec  = t ; 
  timeout.tv_usec = 0 ;

  FD_ZERO ( &fds ) ;
  FD_SET ( faxin , &fds ) ;

  n = select ( FD_SETSIZE , &fds , 0 , 0 , t<0 ? 0 : &timeout ) ;
  if ( n < 0 ) msg("SE2select() failed:") ;

  return n ;
}


/* faxundrflw is called only by the faxgetc() macro when the buffer is
   empty.  t is maximum idle time in seconds before giving up. Returns
   number of characters read or EOF on timeout or errors.  */

int faxundrflw ( int t )
{ 
  int n = faxdata ( t ) ;

  if ( n > 0 )
    if ( ( n = read( faxin , faxibuf , RCVBUFSIZE ) ) < 0 ) 
      msg("SE2fax device read:") ;

  faxiq = ( faxip = faxibuf ) + ( n > 0 ? n : 0 ) ;

  return n > 0 ? n : EOF ;
} 


/* Read a string from the modem.  The string is CR-delimited if it begins
   with 'A' (modem command) or a digit (numeric response), otherwise it's a
   LF-delimited modem response.  Returns null if times out in t seconds */

char *faxgets( char *s , int n , int t )
{
  int c=0 ;
  char *p = s ;

  *p = 0 ;
  while ( --n > 0 && ( c = faxgetc( t ) ) != EOF && ( *p++ = c ) != 
	 ( tolower( *s ) == 'a' || isdigit ( *s ) ? CR : LF )  ) ;

  *p = 0 ;
  return c == EOF ? 0 : s ;
}


/* faxputc() is a macro like putc().  It returns the character written or
   EOF on error. */

#define faxputc(c) ( faxop >= faxoq && faxflush() == EOF ? EOF : \
		    (unsigned char) ( *faxop++ = c ) )


/* Write the fax output buffer to fax output device. Returns 0 or EOF on
   error. */

int faxflush ( )
{ 
  int n=0 ;
  char *p = faxobuf ;

  if ( faxop >= faxoq ) msg("Wfax output buffer overflow") ;

  while ( p < faxop && ( n = write( faxout , p , faxop - p ) ) >= 0 )
    p += n ? n : msg ("W0wrote %d of %d bytes" , n , faxop - p ) ;

  return n >= 0 ? (faxop = faxobuf , 0) : (msg("SEfax device write:") , EOF) ;
} 


/* Send character or string to modem immediately (for commands).  Return
   like putc() and puts(). */

int faxputcnow ( char c ) 
{ 
  return faxputc ( c ) < 0 ? EOF : ( faxflush() ? EOF : c ) ; 
}

int faxputs ( char *s )
{
  int n=0 ;
  while ( s && *s && ( n = faxputcnow ( *s++ ) ) != EOF ) ;
  return n ;
}


/* Fill capability struct from capability string. Returns 0 or 2
   on errors. */

int str2cap ( char *s, cap *c )
{
  int err=0, i=0, n, a[NCAP] ;
  char *p ;

  a[0] = c->vr ;   a[1] = c->br ;    a[2] = c->wd ;    a[3] = c->ln ;
  a[4] = c->df ;   a[5] = c->ec ;    a[6] = c->bf ;    a[7] = c->st ;
  
  for ( i=0, p=s ; ! err && *p && i<NCAP ; p++ )
    if ( isdigit ( *p ) ) a [ i++ ] = *p - '0' ;
    else if ( isspace ( *p ) || *p == ',' ) ;
    else err = msg ("E2invalid character (%c) in (%s)", *p , s ) ;
  
  if ( (n=i) != NCAP ) msg ("W0missing values in capability \"%s\"", s ) ;
  
  for ( i=0 ; i<n ; i++ ) 
    if ( a [ i ] > capmax [ i ] ) 
      err = msg( "E2capability field %d (%d) too large", i+1, a [ i ] ) ;
  
  c->vr = a[0] ;   c->br = a[1] ;    c->wd = a[2] ;    c->ln = a[3] ;
  c->df = a[4] ;   c->ec = a[5] ;    c->bf = a[6] ;    c->st = a[7] ;
  
  return err ;
}



/* Search for a match to the string s in a null-terminated array of
   possible prefix strings pointed to by p.  The first character of each
   prefix string is skipped.  Returns pointer to the table entry or NULL if
   not found.  */

char *strtabmatch ( char **p, char *s )
{
  while ( *p && strncmp ( *p+1 , s , strlen ( *p+1 ) ) ) p++ ;
  return *p ;
}


/* Send command to modem and check responses.  Writes the command s to the
   modem if s is not null.  Reads responses if t is non-zero. Notes and
   processes important responses (FPTS: Page Transfer Status, and FHNG:
   HaNGup status, and FDCS: Current Session capabilities).  Terminates when
   a response is one of the prompts in responses[].  Returns the first
   character of the matching prefix string (e.g. 'O' for OK, 'C' for
   CONNECT, etc.)  or EOF if no such response was received within t
   seconds.  Numeric responses are converted to corresponding letter. */

/* should probably try to repeat the last command if times out and the last
   response was RING */

int cmd ( char *s , int t )
{
  char *buf = lastprompt, *p = "" ;
  int ppr ;
  
  usleep ( cmdpause * 1000 ) ;    /* modems need delay between commands */

  if ( verb > 1 ) msg ( s ? "modem command \"%s\"" : "waiting..." , s ) ;

  if ( s ) { faxputs ( "AT" ) ; faxputs ( s ) ; faxputcnow ( CR ) ; }

  while ( t && ( p = faxgets( buf , CMDBUFSIZE , t ) ) ) {
    if ( verb ) logs ( buf ) ;

    if ( ! strncmp ( buf, "+FDCS:", 6 ) && !str2cap ( buf+6 , &session ) ) {
      msg ( "session capabilities: vr=%d,br=%d,df=%d,st=%d" ,
	   session.vr, session.br, session.df, session.st ) ;
      minlen = padding ? cps[ session.br ] * delay[ session.st ] / 1000 : 0 ;
      if ( minlen ) msg ("padding scan lines to %d bytes" ) ; 
      decimate = local.vr == 1 && session.vr == 0 ;
    }

    if ( sscanf ( buf , "+FPTS: %d", &ppr ) > 0 ) good = ppr & 1 ;

    sscanf ( buf , "+FHNG: %d", &hsc ) ;
    
    if ( ( p = strtabmatch ( (char**) responses , buf ) ) ) break ; 
  }

  return p ? *p : EOF ;
}


/* Test for UUCP lock file & remove stale locks. Returns 0 on null file
   name or if no longer locked, 1 if locked by another pid, 2 on error, 3
   if locked by us. */

int ttlocked ( char *fname )
{
  int err=0 ;
  FILE *f ;
  pid_t pid = 0 ;
  char buf [ FILENAME_MAX ] = "" ;

  if ( fname && *fname == HDBLKFLAG ) fname++ ;

  if ( fname && ( f = fopen ( fname , "r" ) ) ) {
    if ( fread ( buf, sizeof(char), FILENAME_MAX-1, f )  == sizeof(pid_t) || 
	sscanf ( buf , "%d" , &pid ) != 1 ) pid = * (pid_t *) buf ;
    if ( kill ( pid , 0 ) && errno == ESRCH )
      if ( unlink ( fname ) ) err = 
	msg ( "SE2can't remove stale lock %s from pid %d:" , fname , pid ) ;
      else err = 
	msg("0removed stale lock %s from pid %d" , fname , pid ) ;
    else 
      if ( pid != getpid() ) err = 1 ;
      else err = 3 ; 
    fclose ( f ) ;
  }
  return err ;
}

/* Create UUCP (text or binary) lock file.  Returns 0 on null
   file name or if created, 1 if locked by another pid, 2 on
   error, 3 if locked by us. */

int ttlock ( char *fname )
{
  int err=0, dirlen, hdb=0 ;    
  FILE *f=0 ;    
  pid_t pid = getpid ( ) ;
  char *p , buf [ FILENAME_MAX ] = "" ;

  if ( fname && *fname == HDBLKFLAG ) { fname++ ; hdb=1 ; }

  if ( fname && ! ( err = ttlocked ( fname ) ) ) {
    dirlen = ( p = strrchr( fname , '/' ) ) ? p-fname+1 : strlen ( fname ) ;
    sprintf ( buf , "%.*sTMP..%05d" , dirlen , fname , pid ) ;
    if (  ( f = fopen( buf, "w" ) )  && 
	( hdb ? 
	 (fprintf(f, "%10d\n", pid)>0) : fwrite(&pid, sizeof(pid_t), 1, f) )
	) {
      if ( rename ( buf , fname ) == 0 ) chmod ( fname , 0444 ) ;
      else if ( ! ( err = ttlocked ( fname ) ) )
	err = msg ( "SE2can't rename lock file %s to %s:", buf, fname ) ;
    } else {
      err = msg ("SE2can't open/write pre-lock file %s:", buf ) ;
    }
  }
  if ( f ) { fclose ( f ) ; if ( err ) unlink ( buf ) ; }
  return err ;
}


/* Remove lock file.  Returns 0 on null file name, doesn't exist, or was
   removed, 1 if the lock is to another pid, 2 on errors. */

int ttunlock ( char *fname )
{
  int err = 0 ;

  if ( fname && *fname == HDBLKFLAG ) { fname++ ; }

  if ( fname && ( err = ttlocked ( fname ) ) ) {
    if ( err == 1 ) msg ( "Ewon't remove lock %s (not ours)" , fname ) ;
    if ( err == 3 )
      if ( unlink ( fname ) ) err = msg("SE2can't remove lock %s:", fname ) ;
      else err = 0 ;
  }
  return err ;
}


/* Lock all lock files.  Returns 0 if all locks [already] applied, 1 if any
   are locked to other pids, 2 on any errors. */

int lockall ( )
{ 
  int err = 0 ;
  char **p = lkfiles ;
  while ( *p && ! err ) 
    if ( ( err = ttlock ( *p++ ) ) == 3 ) err = 0 ; 
  return err ; 
}

/* Remove all lock files.  Returns 0 if all locks removed, 2 on errors. */

int unlockall ( )
{ 
  int err = 0 ;
  char **p = lkfiles ;
  while ( *p ) if ( ttunlock ( *p++ ) ) err = 2 ; 
  return err ; 
}


/* Resynchronize modem from an unknown state.  If no immediate response,
   try pulsing DTR low (needs &D{2,3,4}), and cancelling data or fax data
   modes.  In each case, discard any responses for 1 or 2 seconds and then
   try command s.  Returns 0 if OK or 4 if no response.  */

int faxsync( char *s )
{
  int err=0, method=0 ;

  while ( ! err ) {
    switch ( method++ ) {
    case 0 : 
      break ;
    case 1 : 
      msg ("sync: trying to drop DTR") ;
      ttymode( CMD_MODE ) ; usleep ( T_DTR*1000 ) ;
      ttymode( HUP_MODE ) ; usleep ( T_DTR*1000 ) ;
      ttymode( CMD_MODE ) ; usleep ( T_DTR*1000 ) ;  
      break ;
    case 2 : 
      msg ("sync: trying data escape") ;
      usleep ( T_ESC*1000 ) ; faxputs ( "+++" ) ; break ;
    case 3 : 
      msg ("sync: trying fax escape") ;
      faxputcnow ( DLE ) ; faxputcnow ( ETX ) ; break ;
    case 4 :
      err = msg("E4modem not responding") ;
      continue ;
    }
    while ( cmd ( 0 , method==1 ? 1 : TO_RESET ) != EOF ) ;
    if ( cmd ( s , TO_RESET ) == OK ) break ;
  }
  return err ;
} 


int end_session ( )
{
  int err = faxsync( "Z" ) ;
  if ( ! err ) ttymode ( OLD_MODE ) ;
  unlockall ( ) ;
  return err ;
} 
    

/* signal handler: hang up and exit */

void onsig ( int sig ) 
{ 
  msg("ETterminating on signal %d",sig) ; 
  end_session ( ) ;
  msg ("done, returning 5") ;
  exit(5) ; 
} 


/* Initialize session.  Try locking and opening fax device until opened or
   get error. Then set tty modes, register signal handler, set modem
   options.  */

int begin_session ( char *faxdev )
{
  int i, err=0 , catch[] = { SIGNALS , 0 } ;

  msg ("waiting for %s ...", faxdev ) ;
  do {
    err = lockall ( ) ;
    
    if ( ! err )
      if ( ! strcmp ( faxdev , "-" ) ) {
	faxin = fileno ( stdin ) ; 
	faxout = fileno ( stdout ) ; 
      } else {
	if ( ( faxin = faxout = open ( faxdev , O_RDWR ) ) < 0 )
	  if ( errno == EBUSY ) err = 1 ; 
	  else err = msg("SE2unable to open fax device\n%s:", faxdev ) ;
      }
    
    if ( err == 1 ) sleep ( LOCKPOLLDELAY ) ;

  } while ( err == 1 ) ;
  
  if ( ! err ) msg ("opened" ) ;

  if ( ! err ) err = ttymode ( CMD_MODE ) ;
  
  for ( i=0 ; ! err && catch [ i ] ; i++ ) 
    if ( signal ( catch [ i ] , onsig ) == SIG_ERR ) 
      err = msg ("SE2can't set signal %d handler:", catch [ i ] ) ;
  
  if ( !err ) err = faxsync("Z") ;

  for ( i=0 ; ! err && i<niopt ; i++ )
    if ( cmd( iopt[i] , TO_INIT ) != OK && ! igniniterr )
      err = msg("E3modem initialization command (%s) failed", iopt[i]) ;

  return err ;
}


/* Check for end-of-line (a 1 following 11 or more zero bits) in T.4
   data. Returns 1 on EOL, 2 on EOL after empty line, 3 on T.4 RTC (6 or
   more successive EOLs).  Initialize before each file by calling with a
   negative value.  */

int tsteol ( int c )
{
  int eol=0 ;
  static int emptylines, ocount, zcount ;

  if ( c < 0 ) emptylines=ocount=zcount=0 ;
  else
    if ( ( c &= 0xff ) ) {
      if ( ( zcount += eolprezero [ c ] ) >= 11 ) { 
	if ( ocount == 1 ) eol = ( ++emptylines >= 5 ) ? 3 : 2 ;
	else { eol = 1 ; emptylines = 0 ; }
	ocount = eolones [ c ] ;
      } else {
	ocount += eolones [ c ] ;
      }
      zcount = eolpostzero [ c ] ;
    } else {
      zcount += 8 ;
    }
  return eol ;
}


/* Send data for one page. Set up serial port and reset end-of-line parser.
   Read characters from file and pad lines if necessary.  If enough
   characters stored, flush buffer and check for CANcel character.
   Bit-reverse characters and escape DLE before storing in buffer.  Send
   RTC if needed.  Send DLE-ETX and return serial port to command mode when
   done. n is bytes in line, m is (approximate) bytes in output buffer */

/* bug: doubles the amount of padding when decimating */

int send_data( FILE *f )
{
  int done=0, err=0, can=0, eol=0, n=0, m=0, bytes=0, lines=0, lastc=0, c, i ;
  int skip=0 ;

  ttymode ( TX_MODE ) ; 
  tsteol ( EOF ) ;

  while ( ! done ) {
    
    if ( ( c = getc ( f ) ) != EOF ) {

      if ( ( eol = tsteol ( c ) ) ) {
	lines++ ; bytes+=n ;
	while ( n < minlen || lastc ) { faxputc ( lastc = 0 ) ; n++ ; }
	if ( ( m += n ) > MINWRITE ) { if ( faxflush() ) done=err=2 ;  m=0 ; }
	n = 0 ;
	if ( eol == 3 ) done = 1 ;
	skip = decimate && ( ~lines & 1 ) ;
      }

      if ( skip ) continue ;

      faxputc ( lastc = c = bitorder [ c ] ) ;
      if ( c == DLE ) faxputc ( DLE ) ;
      n++ ;

      if ( ! m )
	while ( ( c = faxgetc ( 0 ) ) != EOF ) 
	  if ( c == CAN ) done = can = msg("W2transmission CANcellled") ;
	  else msg("W[%02x] received while sending data", c ) ;
	
    } else {
      done = 1 ;
      if ( ferror ( f ) ) err = msg("SE2fax file read:") ;
    }
  }

  if ( eol != 3 ) {
    msg ( "RTC (%d bytes) appended" , T4RTCLEN ) ;
    for ( i=0 ; i<T4RTCLEN ; i++ ) faxputc ( (T4RTC)[i] ) ;
  }

  faxputc ( DLE ) ; faxputc ( ETX ) ; 
  if ( can ) faxputc ( CAN ) ;
  if ( faxflush ( ) ) err = 2 ;

  ttymode ( FAX_MODE ) ; 
  
  if ( cmd ( 0 , TO_B ) != OK )
    err = msg("E3no response after sending data") ; 

  msg ( "sent %d data bytes in %d scan lines" , bytes , lines ) ;

  return err ;
}


/* Receive data. Enable input flow control and send DC2 to start the
   transfer.  Get a character from the modem and check for errors.  Remove
   long runs of zeroes (T.4 FILL). If in DLE escape check for DLE, end of
   data or protocol violation.  Otherwise, test for DLE & write the
   character out.  Check that the output file is still OK.  If not, send
   one CANcel character and wait for protocol to complete.  */

int receive_data ( FILE *f )
{
  int done=0, err=0, dle=0, nulls=0, bytes=0, lines=0, eol=0, c ;

  ttymode ( RX_MODE ) ; 

  tsteol ( EOF ) ;

  if ( startchar ) faxputcnow ( startchar ) ;

  while ( !done ) {
	
    if ( ( c = faxgetc( TO_C ) ) != EOF ) {

      if ( padding && ( nulls = ( c ? 0 : nulls+1 ) ) >= MAXNULLS ) continue ;

      if ( dle ) {
	dle = 0 ;
	if ( c == DLE ) goto isdata ;
	else if ( c == ETX ) done = 1 ;
	else msg("W[%02x] received after DLE", c ) ;
      } else {
	if ( c == DLE ) dle = 1 ;
	else {
	isdata:
	  bytes++ ;
	  if ( ( eol = tsteol ( c ) ) ) lines++ ;
	  putc ( bitorder [ c ] , f ) ;
	}
      }
      
      if ( ! err && ferror ( f ) ) {
	err = msg("SE2fax file write:") ;
	faxputcnow ( CAN ) ;
	msg("Wreceive CANcelled") ;
      } 

    } else
      done = err = msg("E3fax device read timeout") ;
  }

  if ( ! err && eol != 3 ) {
    msg ( "RTC (%d bytes) appended" , T4RTCLEN ) ;
    fwrite ( T4RTC , sizeof ( char ) , T4RTCLEN , f ) ;
  }

  ttymode ( FAX_MODE ) ; 

  if ( cmd ( 0 , TO_B ) != OK )
    err = msg("E3no response after receiving data") ; 
  
  msg ( "received %d data bytes in %d scan lines" , bytes , lines ) ;

  return err ;
}


/* Send a page. Open the image file for binary reading.  Give the data
   transmit command. Transmission begins after DC2 (or XON) is received.
   Sends the data, sends appropriate post-page message, and closes the
   image file.  */

int send_page ( char *fname , int last )
{
  int err=0 , c ;
  FILE *f ;

  if ( ( f = fopen ( fname , "rb" ) ) == 0 ) 
    err = msg("SE2fax file open\n%s:",fname) ; 
  else
    msg("opened fax file %s", fname ) ;

  if ( decimate ) msg ( "Wconverting 198 lpi to 96 lpi" ) ;

  if ( ! err )
    if ( cmd ( "+FDT" , TO_B ) != CONNECT ) 
      err = msg("E3send (+FDT) command failed") ; 
  
  if ( wait2send )
    while ( !err && ( c = faxgetc ( TO_B ) ) != XON && c != DC2 )
      if ( c == EOF ) 
	err = msg("E3timed out waiting to send data") ;
      else 
	msg("W[%02x] received while waiting to send data", c ) ;
  
  if ( ! err ) 
    err = send_data ( f ) ;

  if ( ! err )
    if ( cmd ( last ? "+FET=2" : "+FET=0" , TO_B ) != OK )
      err = msg("E3no post-page-response") ; 
  
  if ( f )
    if ( fclose ( f ) != 0 ) 
      msg("SEfax file close:") ; 
    else 
      msg("Iclosed fax file") ;
  
  return err ; 
}

/* Dial a number and send a fax.  Sends files fname[0...n-1].  */

int send ( char *tel , int n , char *fname[] , char *faxdev )
{
  char c, s [ 128 ] ;
  int err=0 , page , try ;

  msg("sending %d pages to %s using %s", n, tel, faxdev ) ;

  sprintf ( s , "D%.127s" , tel ) ;

  err = begin_session ( faxdev ) ;

  if ( ! err ) {
    ttymode ( FAX_MODE ) ;
    if ( ( c = cmd ( s , TO_A ) ) == OK && hsc < 0 ) 
      msg("connection established" ) ; 
    else if ( c ==  BUSY ) 
      err = msg("W1number is busy" ) ; 
    else 
      err = msg("E2can't establish session" ) ;
  }
  
  for ( page=0 ; !err && page < n ; page++ ) {

    for ( try=0 ; !err && try <= NTXRETRY && hsc < 0 ; try++ ) {
      good = 1 ;
      err = send_page ( fname[page] , page >= n-1 ) ;
      if ( good ) break ;
    }

    if ( ! err && ! good ) err = msg("E2too many page send retries" ) ;
    if ( hsc > 0 ) err = msg("E2abnormal call termination (code %d)", hsc ) ;

  } 
  if ( err < 4 ) end_session( ) ;

  return err ;
}


/* Receive a fax page.  Open file for binary write.  Send fax data receive
   command.  If response is OK, then no more pages are coming.  If it's
   CONNECT receive the data for one page.  Returns 0 if a page was
   received, -1 if no more pages, 2 or more for errors.  */

int receive_page ( char *fname )
{
  int err=0, c ;
  FILE *f ;

  if ( ( f = fopen ( fname , "wb" ) ) == 0 )
    err = msg("SE2fax file open\n%s:",fname) ; 
  else
    msg("opened file %s", fname ) ;
  
  if ( ! err )
    if ( ( c = cmd ( "+FDR" , TO_B ) ) != CONNECT )
      if ( c == OK )
	err = -1 ;
      else 
	err = msg("E3receive (+FDR) command failed") ;
  
  if ( ! err ) 
    err = receive_data ( f ) ;
  
  if ( f ) 
    if ( fclose ( f ) != 0 ) 
      msg("SEfax file close:") ; 
    else 
      msg("closed fax file") ;

  if ( ! err ) msg ( "received -> %s ", fname ) ;  /* mark name for scripts */

  if ( err == -1 ) {	    /* remove the (empty) last file */
    unlink ( fname ) ;
    msg ( "removed %s", fname ) ;
  }

  return err ;
}

/* Receive a fax.  Open modem device and initialize it.  Remove locks if
   sharing device with outgoing calls.  If waiting for call, wait for modem
   activity, else answer phone.  Re-lock if necessary. Generate base file
   name if it's missing.  Receive one file per page.  Make file name by
   appending a 3-digit number to base string. */

int receive ( char *fmt , char *faxdev )
{
  enum connectmode { NONE, DATA, FAX } ; 
  enum connectmode mode=NONE ;
  int c=0, err, page ;
  char fname [ FILENAME_MAX + 1 ] , base [ FILENAME_MAX + 1 ] ;
  time_t t = time(0)  ;

  err = begin_session ( faxdev ) ;

  if ( ! err && share ) err = unlockall ( ) ;

  if ( ! err && waitforcall ) {
    msg ( "waiting for activity on %s", faxdev ) ;
    faxdata ( -1 ) ;
    msg ("activity detected") ;
  }
  
  if ( ! err && share ) err = lockall ( ) ;

  if ( ! err ) {
    if ( datafirst && *getty )
      if ( cmd ( ( waitforcall ? 0 : "A" ) , TO_DATAF ) == CONNECT ) 
	mode = DATA ;
      else { 
	int i ;   /* abort data answer mode & set fax mode */
	for ( i=0 ; i<3 && cmd ( "+FCLASS=2" , TO_RESET ) != OK ; i++ ) ; 
      }
    if ( mode != DATA )
      if ( ( c = cmd ( ! waitforcall || ( datafirst && *getty ) ? "A" : 0 , 
		      TO_A ) )  == OK && hsc < 0 )
	mode = FAX ;
      else
	if ( ( c == CONNECT ) && *getty ) mode = DATA ;
  }

  if ( ! err )
    switch ( mode ) {
    case DATA : {
        char *p, buf [ MAXGETTY ] ;
	msg ("data call answered") ;
	if ( !strtok(lastprompt, " ") || !(p=strtok(0, "\n\r\t "))) p="300";
	sprintf ( buf , getty , p,p,p,p, p,p,p,p, p,p,p,p ) ;
	msg ( "exec'ing /bin/sh -c \"%s\"" , buf ) ;
	execl ( "/bin/sh" , "/bin/sh" , "-c" , buf , (char*) 0 ) ; 
	err = msg ( "SE2exec failed:" , buf ) ;
	break ; }
      case FAX:
	msg ("fax call answered") ;
	err = ttymode ( FAX_MODE ) ;
	break ;
      case NONE:
	err = msg("E3unable to answer call") ;
	break ;
      }

  strftime( base, FILENAME_MAX, *fmt ? fmt : FNAMFT , localtime ( &t ) ) ;
  
  if ( ! err ) msg("receiving %s.nnn from %s", base, faxdev ) ;
  
  for ( page=1 ; ! err && hsc < 0 ; page++ ) {
    sprintf ( fname, "%.*s.%03d", FILENAME_MAX - 4, base, page ) ;
    err = receive_page ( fname )  ;
  } 
  
  if ( err != 1 && err < 4 ) end_session( ) ;   /* if not locked */
  
  if ( err == -1 ) err = 0 ;  /* normal last page, return 0 */
  if ( err ==  5 ) err = 0 ;  /* data call, return 0 */
  if ( hsc > 0 ) err = msg("E2abnormal call termination (code %d)", hsc ) ;

  return err  ;
}


/* Simple (one option per argument) version of getopt(3). */

int optind = 1 ;
char *optarg ;

int nextopt( int argc, char **argv, char *args )
{
  char *a, *p ;

  if ( optind >= argc || *(a = argv[optind]) != '-' ) return -1 ;
  optind++ ;

  if ( ! *(a+1) || ! ( p = strchr ( args , *(a+1) ) ) )
    return msg ("Eunknown option (%s)" , a ) , '?' ; 

  if ( *(p+1) != ':' ) optarg = 0 ;
  else
    if ( *(a+2) ) optarg = a+2 ;
    else
      if ( optind >= argc ) return msg ( "Eno argument for (%s)", a ) , '?' ;
      else optarg = argv [ optind++ ] ;
  return *(a+1) ;
}


/* Fax send/receive program for class 2 fax modems. Print version,
   initialize bit reversal and end-of-line parsing tables and process
   arguments.  Returns 0 on success, 1 if busy, 2 for errors, 3 for
   protocol errors, and 4 if no modem response. */

int main( int argc, char **argv)
{
  int i, err=0 , c, nlocks=0 ;
  char *faxdev = FAXDEV ;
  char capinit [ CMDBUFSIZE ] , idinit [ CMDBUFSIZE ] ;

  argv0 = argv[0] ;

  msg ( Version " (" __DATE__ " " __TIME__ ") starts" ) ;

  for ( i=0 ; i<256 ; i++ ) normalbits [ reversebits [ i ] = i ] = 
    ( i& 1 ? 128:0 ) | ( i& 2 ? 64:0 ) | ( i& 4 ? 32:0 ) | ( i&  8 ? 16:0 ) |
    ( i&16 ?   8:0 ) | ( i&32 ?  4:0 ) | ( i&64 ?  2:0 ) | ( i&128 ?  1:0 ) ;
  
  for ( i = 0 ; i<256 ; i++ ) {
    unsigned char m ;
    for ( m=0x80 ; m ; m>>=1 ) if ( i & m ) eolones[i] ++ ; 
    for ( m=0x80 ; ~i & m && m ; m>>=1 )    eolprezero[i] ++ ; 
    for ( m=0x01 ; ~i & m && m ; m<<=1 )    eolpostzero[i] ++ ; 
  }

  while (!err && (c=nextopt(argc, argv, "b:c:d:g:i:l:o:r:st:vwx:") ) != -1) {
    
    switch (c) {
    case 'b':
      if ( cbaud ( optarg ) != EOF ) cmdcbaud = cbaud ( optarg ) ; 
      break ;
    case 'c': 
      err = str2cap ( optarg , &local ) ;
      sprintf ( capinit , "+FDCC=%.*s" , CMDBUFSIZE-7, optarg ) ;
      if ( !err && !c1 ) { optarg = capinit ; goto addopt ; }
      break ;
    case 'l': 
      localid = optarg ;
      sprintf ( idinit , "+FLID=\"%.*s\"" , CMDBUFSIZE-7, optarg ) ;
      if ( !c1 ) { optarg = idinit ; goto addopt ; }
      break ;
    case 'i': 
    addopt:
      if ( niopt < MAXIOPT ) iopt [ niopt++ ] = optarg ;
      else err = msg("E2too many modem init options"); 
      break ;
    case 'd': faxdev = optarg ; break ;
    case 'g': getty = optarg ; break ;
    case 'o': 
      for ( ; *optarg ; optarg++ ) 
	switch ( *optarg ) {
/*	case '1' : c1 = 1 ; break ; */
	case '2' : c1 = 0 ; break ;
	case 'a' : datafirst = 1 ;  break ;
	case 'b' : hispeed = 1 ;  break ;
	case 'e' : igniniterr = igniniterr ? 0 : 1 ;  break ;
	case 'f' : flowctlrcv = 0 ;  break ;
	case 'h' : rtscts = 1 ; break ;
	case 'p' : padding = 1 ; break ;
	case 'r' : bitorder = reversebits ; break ;
	case 's' : stopbits = 1 ; break ;
	case 'w' : wait2send = 0 ;   break ;
	case 'x' : startchar = XON ; break ;
	case 'z' : cmdpause += T_CMD ; break ;
	 default : msg ( "Wunrecognized protocol option (%c)", *optarg ) ; 
	}
      break ;
    case 'r': err = receive ( optarg , faxdev ) ; break;
    case 's': share = 1 ; break;
    case 't': err = send ( optarg, argc-optind, argv+optind, faxdev ) ; break;
    case 'v': 
      verb++ ; 
      for ( i=0 ; verb==2 && i<argc ; i++ ) msg("argv[%d]=%s", i, argv[i]) ; 
      break ;
    case 'w': waitforcall = 1 ; break ;
    case 'x': 
      if ( nlocks < MAXLKFILE ) lkfiles [ nlocks++ ] = optarg ; 
      else err = msg( "E2too many lock files" ) ; 
      break ;
    default : fprintf ( stderr, Usage, argv0 ) ; err = 2 ; break ;
    }
  }

  msg("done, returning %d", err ) ;
  return err ;
}
